/* -*-C-*-
 *
 * $Revision: 1.2 $
 *   $Author: mquinn $
 *     $Date: 1999/01/26 14:09:38 $
 *
 * Copyright (c) 1996 Advanced RISC Machines Limited.
 * All Rights Reserved.
 *
 *   Project: ANGEL
 *
 *     Title: Serial device driver for EBSA-110 board's PC87311 devices.
 */

/*
 * Provide a single serial device.  Since we're only providing a single
 * device we don't need to worry about device id's and idents.  But these
 * could be used to provide for two or more devices via the one driver.
 */

#include "serial.h"             /* header for this file */

#include "devdriv.h"            /* device driver support */
#include "platform.h"
#include "angel285.h"           /* description of serial hardware */

#include "params.h"             /* parameter structures and utilities */
#include "rxtx.h"               /* rx/tx packet engines */
#include "support.h"            /* __rt_[en/dis]ableIRQ() */
#include "logging.h"            /* ERROR(), etc. */
#include "serlock.h"            /* serialisation, etc. */
#include "serring.h"            /* interface to high-level driver */
#include "banner.h"             /* Copy banner stuff to serial */

/* General purpose constants, macros, enums, typedefs */

#define DEFBAUD        9600        /* default baud rate */

/* Forward declarations */

static void serial_ControlTx(DeviceID devid);
static void serial_ControlRx(DeviceID devid);
static DevError serial_Control(DeviceID devid, 
                               DeviceControl op, void *arg);
static void serial_KickStartFn(DeviceID devid);

static DevError serial_set_params(DeviceID devid, const ParameterConfig *conf);


/*
 * The set of parameter options supported by the device
 */
static unsigned int baud_options[] = { 1200, 2400, 9600, 19200, 38400, 57600, 115200 };

static ParameterList param_list[] = {
    {
        AP_BAUD_RATE,
        sizeof(baud_options) / sizeof(baud_options[0]),
        baud_options
    }
};

/* 
 *The default parameter config for the device
 */
static Parameter param_default[] = {
    { AP_BAUD_RATE, 9600 }
};

/* the state shared between the interrupt handler and the deferred processor */
static struct RingBuffer serial_rx_ring;
static struct RingBuffer serial_tx_ring;

#if !defined(RAW_EBS_SERIAL) || (RAW_EBS_SERIAL == 0) /* Angel/shared */

/* the configuration needed by the TX and RX engines */

#define SERIAL_FC_SET  ((1<<serial_XON)|(1<<serial_XOFF))
#define SERIAL_CTL_SET ((1<<serial_STX)|(1<<serial_ETX)|(1<<serial_ESC))
#define SERIAL_ESC_SET (SERIAL_FC_SET|SERIAL_CTL_SET)

static const struct re_config engine_config = {
    serial_STX, serial_ETX, serial_ESC, /* self-explanatory?           */
    SERIAL_FC_SET,                    /* set of flow-control characters  */
    SERIAL_ESC_SET,                  /* set of characters to be escaped */
    serpkt_flow_control, (void *)DI_SERIAL,    /* what to do with FC chars */
    angel_DD_RxEng_BufferAlloc, (void *)DI_SERIAL   /* how to get a buffer */
};

/* the state of the rx engine */
static struct re_state rx_engine_state;

/* the state of the tx engine */
static struct te_state tx_engine_state;

/* packet for actual rx in progress */
static struct data_packet  rx_packet;

/* the current write packet */
static struct data_packet  tx_packet;

/* collected TX and RX engine state */
static RxTxState serial_rx_tx_state = 
{
    &engine_config,
    &rx_engine_state,
    &tx_engine_state,
    &rx_packet,
    &tx_packet,
    1 /* rx_flow */
};

#else  /* raw */

static RawState serial_raw_state;

#endif /* ... else raw ... */


/*
 * The control functions and interface
 */
static const SerialControl serial_ctrl = 
{
#if !defined(RAW_EBS_SERIAL) || (RAW_EBS_SERIAL == 0)
    &serial_rx_tx_state,
    NULL,
    serpkt_int_tx_processing,
    serpkt_int_rx_processing,
#else
    NULL,
    &serial_raw_state,
    serraw_int_tx_processing,
    serraw_int_rx_processing,
#endif
    SER_DEV_IDENT_0,
    &serial_tx_ring,
    &serial_rx_ring,
    serial_ControlTx,
    serial_ControlRx,
    serial_Control,
    serial_KickStartFn
};

/* Publically-accessible globals */

/*
 * This is the device table entry for this device
 */
const struct angel_DeviceEntry angel_SerialDevice = {
#if !defined(RAW_EBS_SERIAL) || (RAW_EBS_SERIAL == 0)
    DT_ANGEL,
    {
        serpkt_AsyncWrite,
        serpkt_RegisterRead
    },
    serpkt_Control, 
#else
    DT_RAW,
    {
        /* nasty but necessary casts, as can only statically initialise
         * the first member of a union
         */
        (angel_DeviceWriteFn)       serraw_Write,
        (angel_DeviceRegisterReadFn)serraw_Read
    },
    serraw_Control,
#endif
    &serial_ctrl,
    { sizeof(param_list) / sizeof(param_list[0]), param_list },
    { sizeof(param_default) / sizeof(param_default[0]), param_default }
};

/* Private globals */

/* a copy of the Interrupt Mask Register (since it is Write-Only) */
static word IER_copy = 0;

/* macros for reading and writing registers */
#define WriteReg(b, p)  outb(b, (void *)(HOST_COMPORT + p) )
#define ReadReg(p)      inb( (void *)(HOST_COMPORT + p) )


/* -------------------------- DEBUG code -------------------------- */

#ifndef DEBUG_EB285
#define DEBUG_EB285 0
#endif
#if DEBUG_EB285

/*  DEBUG_EB285 is also defined in the Makefile - change it there */

static int com2_init = 1 ;
int xinb(void *address);
int xoutb(unsigned char value, int address);

void putchar(byte Ch)
{
  unsigned int Status;

  if (com2_init) {
    com2_init = 0 ;

    /* set the baud rate to 115200 and a word length of 8 */
    xoutb(LCR_DLAB + 0x3, UART_LCR + DEBUG_COMPORT) ;
    xoutb(1, UART_DLLSB + DEBUG_COMPORT) ;
    xoutb(0, UART_DLMSB + DEBUG_COMPORT) ;
    xoutb(3, UART_LCR + DEBUG_COMPORT) ;

    /* disable the fifo control */
    xoutb(0x8F, UART_FIFO_CONTROL + DEBUG_COMPORT) ;

    /* disable the modem control */
    xoutb(0xF, UART_MCR + DEBUG_COMPORT) ;

    /* clear out any errors by reading the line status register */
    Status = xinb((void *)(UART_LSR + DEBUG_COMPORT)) ;
  } 

  do {
    Status = xinb((void *)(DEBUG_COMPORT+UART_LSR)) ;
  } while(!(Status & LSR_THRE));  /* wait until ready */
  
  xoutb(Ch, DEBUG_COMPORT + UART_TX) ;

  if (Ch == '\n') {
    do {
      Status = xinb((void *)(DEBUG_COMPORT+UART_LSR)) ;
    } while(!(Status & LSR_THRE));  /* wait until ready */
  
    xoutb('\r', DEBUG_COMPORT + UART_TX) ; 
  }
} /* putchar */

void putstring(char *string) 
{
  while (*string != '\0') {
    putchar(*string) ;
    string++ ;
  }
}

void puthex(unsigned int hex) 
{
  unsigned char c ;

  c = hex & 0x00f0 ;
  c = c >> 4 ;
  if (c > 9) c = c - 10 + 'A' ;
  else c += '0' ;
  putchar(c) ;

  c = hex & 0x000f ;
  if (c > 9) c = c - 10 + 'A' ;
  else c += '0' ;
  putchar(c) ;
}

int isalnum(unsigned char c) 
{
  if ((c >= 'a') && (c <= 'z')) return 1 ;
  if ((c >= 'A') && (c <= 'Z')) return 1 ;
  if ((c >= '0') && (c <= '9')) return 1 ;
  return 0 ;
}

int isprintable(unsigned char c) 
{
  if ((c >= ' ') && (c <= '~')) return 1 ;
  return 0 ;
}

void put(unsigned int v)
{
  putchar('|') ;
  puthex((v >> 24) & 0xff) ;
  puthex((v >> 16) & 0xff) ;
  puthex((v >> 8) & 0xff) ;
  puthex(v & 0xff) ;
}

void puttx(unsigned char c)
{
  putchar((c == 0x1c) ? '[' : ' ') ;
  if (isprintable(c)) {
    putchar(c) ;
  }
  else 
    puthex(c) ;
  if(c == 0x1d)
    putchar('\n') ;
}

void putrx(unsigned char c)
{
  putchar((c == 0x1c) ? '<' : ' ') ;
  if (isprintable(c)) {
    putchar(c) ;
  }
  else 
    puthex(c) ;
  if(c == 0x1d)
    putchar('\n') ;
}
#endif

/* macros/routines to get and put serial characters */
#define _GetChar() GET_CHAR(HOST_COMPORT)

#pragma no_check_stack
static void _PutChar(const char c) 
{
  int status ;

  /* wait until we can transmit */
  do {
    status = GET_STATUS(HOST_COMPORT) ;
  } while (!TX_READY(status)) ;

  PUT_CHAR(HOST_COMPORT, c) ;

#if DEBUG_EB285 && DEBUG_EB285 > 1
  puttx(c) ;
#endif
}
#pragma check_stack


#pragma no_check_stack

/*
        Had problems sending banner over serial
        when default device is ethernet. Will investigate a proper
        fix with ARM for next release. <GJS>
*/
static void squirt_banner(void)
{
#if (ETHERNET_SUPPORTED == 1)
static const char angel_banner[] = ANGEL_BANNER;
int i=0;
        /* Can't print MAC cos ethernet isn't inititailised */
        while (i < sizeof(angel_banner)) {
                _PutChar(angel_banner[i++]);
        };
        _PutChar('\r');
                                  
#endif
}


/*
 *  Function: serial_KickStartFn
 *   Purpose: Kick-start tx by sending first character
 *
 *    Params:
 *       Input: devid   device ID of the driver
 *
 *   Returns: Nothing
 */
static void serial_KickStartFn(DeviceID devid)
{
    char c;

    IGNORE(devid);

#if DEBUG_EB285 && DEBUG_EB285 > 4
        putstring("Kick\n");
#endif
    c = ringBufGetChar( &serial_tx_ring );
    _PutChar(c) ;
}
#pragma check_stack

/* macros to control serial chip */
#define _ResetError() 



/*
 * enable async transmission if not already enabled
 * NB caller should ensure protection against interrupt
 */
#pragma no_check_stack

static void serial_ControlTx(DeviceID devid)
{
    unsigned tx_status = angel_DeviceStatus[DI_SERIAL] & SER_TX_MASK;

    IGNORE(devid);
    /*
     * We enable tx whenever we want to send a flow control char,
     * OR when we are sending a packet and we have not been XOFF'ed
     */

    if (    ( tx_status & SER_TX_FLOW_CONTROL ) 
         || ( tx_status == SER_TX_DATA ) ) {
      /* tx interrupt should be enabled */
#if DEBUG_EB285 > 2
        putstring(" EnT\n") ;
#endif
      IER_copy |= TX_ENABLE ;
      ENABLE_INT(IER_copy, TX_ENABLE, HOST_COMPORT) ;
      angel_DeviceStatus[DI_SERIAL] |= SER_TX_IRQ_EN;
    } else {
      /* tx interrupt should be disabled */
      IER_copy &= ~TX_ENABLE ;
      DISABLE_INT(IER_copy, TX_ENABLE, HOST_COMPORT) ;
      angel_DeviceStatus[DI_SERIAL] &= ~SER_TX_IRQ_EN;
#if DEBUG_EB285 > 2
        putstring(" DisT\n") ;
#endif
    }
#if DEBUG_EB285 > 5
        puthex(IER_copy) ;
        putchar('\n') ;
#endif
}

#pragma check_stack


/*
 * control rx interrupt mask according to flow control and read interest
 * NB caller should ensure protection against interrupt
 */
static void serial_ControlRx( DeviceID devid )
{
    IGNORE(devid);

    if (    ! (angel_DeviceStatus[DI_SERIAL] & SER_RX_DISABLED)
         && ( angel_DeviceStatus[DI_SERIAL] & DEV_READ_BUSY_MASK ) )
    {
        /* rx interrupt should be enabled */
#if DEBUG_EB285 > 2
        putstring(" EnR\n") ;
#endif
      IER_copy |= RX_ENABLE ;
      ENABLE_INT(IER_copy, RX_ENABLE, HOST_COMPORT) ;
      angel_DeviceStatus[DI_SERIAL] |= SER_RX_IRQ_EN;
    } else {
      /* rx interrupt should be disabled */
#if DEBUG_EB285 > 2
        putstring(" DisR\n") ;
#endif
      IER_copy &= ~RX_ENABLE ;
      DISABLE_INT(IER_copy, RX_ENABLE, HOST_COMPORT) ;
      angel_DeviceStatus[DI_SERIAL] &= ~SER_RX_IRQ_EN;
    }
#if DEBUG_EB285 > 5
        puthex(IER_copy) ;
        putchar('\n') ;
#endif
}


/*
 * wait for tx to drain completely
 */
static void serial_tx_drain( void )
{
        /* Wait for message to move over to UART */
        while ( angel_DeviceStatus[DI_SERIAL] & SER_TX_EOD )
                ;

        /* Now wait for UART to drain */
        while( ! TX_EMPTY(HOST_COMPORT) )
                ;
}


/* Routine to write a byte to the given address */
int outb(unsigned char value, void *address)
{
        IO_WRITE(address, value) ;
        return (0) ;
}

/* Routine to read a byte from the given address */
int inb(void *address)
{
        return (IO_READ(address)) ;
}

/* Routine to write a byte to the given address in XBus space */
int xoutb(unsigned char value, int address)
{
        XIO_WRITE(address, value) ;
        return (0) ;
}

/* Routine to read a byte from the given address in XBus space */
int xinb(void *address)
{
        return (XIO_READ(address)) ;
}

/*
 * Harware reset of UART
 */
static DevError ResetDriver( DeviceID devid )
{
    serial_tx_drain();

    return ( serial_set_params( devid, &angel_SerialDevice.default_config ) );
}


/*
 * Initialisation control operation
 */
static DevError serial_init(DeviceID devid)
{
   DevError err;

   /*
    * do one-time start-up initialisation
    * (for this device, just a device reset)
    */
   Angel_ExitToUSR();      /* because DeviceControl requires it */

#if defined(MINIMAL_ANGEL) && MINIMAL_ANGEL != 0
   err=Angel_RawDeviceControl(devid, DC_RESET, NULL);
#else
   err=angel_DeviceControl(devid, DC_RESET, NULL);
#endif

        squirt_banner();
   Angel_EnterSVC();

   return err;
}


extern void     SetLEDs(int) ;
/*
 * Reset control operation
 */
static DevError serial_reset(DeviceID devid)
{

#if DEBUG_EB285
        putstring("\nUart Reset\n") ;
#endif
   (void)ResetDriver(devid);        /* low_level reset */

   /* reset private flags (LEAVE lowest 8 bits alone) */
   angel_DeviceStatus[devid] &= 0xFF;

   /* turn interrupts on/off for rx and tx */
   serial_ControlRx(devid);
   serial_ControlTx(devid);

   /* Turn on the master interrupt enable bit... */
   ENABLE_INTERRUPT(HOST_COMPORT) ;

   return DE_OKAY;
}


/*
 * Receive Mode control operation
 */
static DevError serial_recv_mode(DeviceID devid, DevRecvMode mode)
{
    DevError ret_code = DE_OKAY;

    if ( mode == DR_DISABLE )
    {
        if ( ! (angel_DeviceStatus[DI_SERIAL] & SER_RX_DISABLED) )
        {
            /* disable reception and schedule send of XOFF */
            angel_DeviceStatus[DI_SERIAL] |= (SER_RX_DISABLED | SER_TX_FLOW_CONTROL);
            serial_ControlRx(devid);
            serial_ControlTx(devid);
        }
    }
    else if ( mode == DR_ENABLE )
    {
        if ( angel_DeviceStatus[DI_SERIAL] & SER_RX_DISABLED )
        {
            /* enable reception and schedule send of XON */
            angel_DeviceStatus[DI_SERIAL] &= ~SER_RX_DISABLED;
            angel_DeviceStatus[DI_SERIAL] |= SER_TX_FLOW_CONTROL;
            serial_ControlRx(devid);
            serial_ControlTx(devid);
        }
    }
    else
    {
        ret_code = DE_INVAL;
    }
    
    return ret_code;
}


/*
 * Set Speed control operation
 */
static DevError serial_set_params(DeviceID devid, const ParameterConfig *conf)
{
word    speed;
word    baud_value;

        IGNORE(devid);

        if ( ! Angel_FindParam( AP_BAUD_RATE, conf, &speed ) ) {
                return DE_BAD_OP ;
#if 0
                ERROR( "speed not found" );
#endif
        }

        switch (speed) {
                case   1200: baud_value = BAUD_1200;   break;
                case   2400: baud_value = BAUD_2400;   break;
                case   9600: baud_value = BAUD_9600;   break;
                case  19200: baud_value = BAUD_19200;  break;
                case  38400: baud_value = BAUD_38400;  break;
                case  57600: baud_value = BAUD_57600;  break;
                case 115200: baud_value = BAUD_115200; break;
                default:     
                        return DE_BAD_OP;
        }

#if DEBUG_EB285
        putstring("\nRate ") ;
        puthex(baud_value) ;
#endif

        serial_tx_drain();

#if UARTS == 0

        /* first, disable everything */
        IO_WRITE(HOST_COMPORT + AMBA_UARTCON, 0x0) ;

        /* Set baud rate */
        IO_WRITE(HOST_COMPORT + AMBA_L_UBRLCR, (baud_value & 0xff) ) ;
        IO_WRITE(HOST_COMPORT + AMBA_M_UBRLCR, ((baud_value & 0xf00) >> 8)) ;

        /* ----------v----------v----------v----------v---------- */
        /* NOTE: MUST BE WRITTEN LAST (AFTER M_UBRLCR & L_UBRLCR) */
        /* ----------^----------^----------^----------^---------- */
        /* set the UART to be 8 bits, 1 stop bit, no parity, fifo disabled */
        IO_WRITE(HOST_COMPORT + AMBA_H_UBRLCR, AMBA_H_UBRLCR_DSS_8) ;

        /* finally, enable the uart */
        IO_WRITE(HOST_COMPORT + AMBA_UARTCON, AMBA_UARTCON_UE) ;

#else

        /* disable rx, tx and error interrupts during reset */
        WriteReg(0, UART_IER) ;
        /* set the baud rate and a word length of 8 */
        WriteReg(LCR_DLAB, UART_LCR) ;
        WriteReg(baud_value, UART_DLLSB) ;
        WriteReg(0, UART_DLMSB) ;
        WriteReg(0x3, UART_LCR) ;

        /* disable the fifo control */
        WriteReg(0, UART_FIFO_CONTROL) ;

        /* disable the modem control */
        WriteReg(0xF, UART_MCR) ;

        /* clear out any errors by reading the line status register */
        ReadReg(UART_LSR) ;

        WriteReg(IER_copy, UART_IER) ;

#endif

#if DEBUG_EB285
        putchar('\n') ;
#endif

        return DE_OKAY;
}


/* 
 * implementation of device control for serial driver 
 *
 * devid is guaranteed to be in range but op is not checked as
 * individual devices can extend the range of operations
 */
static DevError serial_Control(DeviceID devid, 
                               DeviceControl op, void *arg)
{
    DevError ret_code;

    /* we only implement the standard ones here */
    switch( op )
    {
       case DC_INIT:
           ret_code = serial_init(devid);
           break;

       case DC_RESET:
           ret_code = serial_reset(devid);
           break;

       case DC_RECEIVE_MODE:  
           ret_code = serial_recv_mode(devid, (DevRecvMode)((int)arg));
           break;

       case DC_SET_PARAMS:     
            ret_code = serial_set_params(devid, (const ParameterConfig *)arg);
           break;

       default:
           ret_code = DE_BAD_OP;
           break;
    }

    return ret_code;
}


/* 
 * interrupt handler for serial driver
 */
#pragma no_check_stack

void angel_SerialIntHandler( unsigned ident, unsigned data,
                             unsigned empty_stack )
{
    word      IntStatus;
    word      IntSource ;

    IGNORE( ident );
    IGNORE( data );
#if defined(MINIMAL_ANGEL) && MINIMAL_ANGEL != 0
    IGNORE( empty_stack );
#endif


    /*
     * This could be a rx or tx interrupt, or an error.
     */

    IntSource = READ_INTERRUPT(HOST_COMPORT) ;

#if DEBUG_EB285 && DEBUG_EB285 > 5
        putstring("IIR(COM1)=0X") ; puthex(IntSource) ; putchar('\n') ;
#endif

    if (IntSource & NO_INTERRUPT) {
      /* no interrupt pending */
#if DEBUG_EB285
        puthex(IntSource) ;
        putstring(" Nu") ;
#endif
      return ;
    }

    
    if (IntSource & RX_INTERRUPT) {
      bool      queue_rx_deferred = FALSE;

      /*
       * Data Ready interrupt
       */
      IntStatus = GET_STATUS(HOST_COMPORT) ;
      if ( RX_DATA(IntStatus) ) {
        do {
        /*
         * For an rx interrupt, place the character into a ring-buffer
         * and queue processing for after we have re-enabled interrpts.
         */
      
          char rx_ch = _GetChar(); /* get a char from UART */
#if DEBUG_EB285 && DEBUG_EB285 > 1
          putrx(rx_ch) ;
#endif
          
          if ( ! ringBufFull( &serial_rx_ring ) ) {
            RawState *const raw_state = serial_ctrl.raw_state;

            ringBufPutChar( &serial_rx_ring, rx_ch );
            if (raw_state != NULL && raw_state->rx_to_go > 0);
              --(raw_state->rx_to_go);

            if (    (raw_state == NULL && rx_ch == serial_ETX)
                    || (raw_state != NULL && raw_state->rx_to_go == 0
                                          && raw_state->rx_data != NULL)
                    || (ringBufCount( &serial_rx_ring ) >= RING_HIGH_WATER_MARK) )
              /* note that we need to do deferred rx processing */
                queue_rx_deferred = TRUE;
          } else {
#if DEBUG_EB285
            putstring(" RO") ;
#endif
#if 0
            WARN( "rx ring buffer overflow" );
#endif
            queue_rx_deferred = TRUE ;
            break ;
          }
          IntStatus = GET_STATUS(HOST_COMPORT) ;
        } while ( RX_DATA(IntStatus) ) ;
      } /* end RX_DATA */
      else {
        /* A receive interrupt with no Data Ready flag! */
#if DEBUG_EB285
        puthex(IntStatus) ;
        putstring(" oR") ;
#endif
      }
        
      /* see if we need to do deferred rx processing */
      if ( queue_rx_deferred && ! (angel_DeviceStatus[DI_SERIAL] & SER_RX_QUEUED) ) {
        angel_DeviceStatus[DI_SERIAL] |= SER_RX_QUEUED;
#if !defined(MINIMAL_ANGEL) || MINIMAL_ANGEL == 0
        Angel_SerialiseTask( 0,
                             serial_ctrl.rx_processing, 
                             (void *)DI_SERIAL,
                             empty_stack,
                             &Angel_GlobalRegBlock[RB_Interrupted]);
#else
        serial_ctrl.rx_processing((void *)DI_SERIAL);
#endif
      }
    } /* IntSource & RX_INTERRUPT */

    else if (IntSource & TX_INTERRUPT) {
      bool      queue_tx_deferred = FALSE;

      /*
       * Ready to transmit interrupt
       */
      IntStatus = GET_STATUS(HOST_COMPORT) ;
      if ( TX_READY(IntStatus) ) {
        do {
          /*
           * For a tx interrupt, get the next character to transmit and send
           * it to the device.
           */
          
          /* do we need to send a flow control character? */
          if ( angel_DeviceStatus[DI_SERIAL] & SER_TX_FLOW_CONTROL ) {
              bool rx_disabled = angel_DeviceStatus[DI_SERIAL] & SER_RX_DISABLED;
              
              /* send a flow control char to UART */
              _PutChar(rx_disabled ? serial_XOFF : serial_XON );

              angel_DeviceStatus[DI_SERIAL] &= ~SER_TX_FLOW_CONTROL;
          }
          else if ( ringBufNotEmpty( &serial_tx_ring ) ) {
              /* send a packet char to UART */
              _PutChar(ringBufGetChar( &serial_tx_ring ) );
          }

          /*
           * if we have reached to low-water mark in the
           * ring buffer then it is time to ask for a refill,
           * as long as there is more of the packet to come
           */
          if (   (ringBufCount( &serial_tx_ring ) < RING_LOW_WATER_MARK)
                 && ! (angel_DeviceStatus[DI_SERIAL] & SER_TX_EOD)   )
            queue_tx_deferred = TRUE;

          /*
           * is it time to give up yet?
           */
          if ( ringBufEmpty( &serial_tx_ring ) ) {
              /*
               * if the end of the packet has been sent,
               * then queue deferred processing of this fact
               */
#if DEBUG_EB285 && DEBUG_EB285 > 4
        putstring("Done\n");
#endif
              if ( angel_DeviceStatus[DI_SERIAL] & SER_TX_EOD ) {
                  queue_tx_deferred = TRUE;
                  
                  /* disable Tx if necessary */
                  angel_DeviceStatus[DI_SERIAL] &= ~SER_TX_DATA;
                  serial_ControlTx(DI_SERIAL);
              }
              else {
                  /*
                   * We need to set a flag to tell fill_tx_ring
                   * that it needs to restart character Tx.
                   */
                  angel_DeviceStatus[DI_SERIAL] |= SER_TX_KICKSTART;
              }

              /* exit the loop, no more to transmit */
              break ;
            }
          IntStatus = GET_STATUS(HOST_COMPORT) ;
        } while ( TX_READY(IntStatus) ) ;
      } 
      else {
        /* A transmit interrupt with no Transmit Holding Register flag! */
#if DEBUG_EB285
        puthex(IntStatus) ;
        putstring(" oT") ;
#endif
      }

      /* If we got here then we may need to queue deferred tx processing */
      if ( queue_tx_deferred &&
           ! (angel_DeviceStatus[DI_SERIAL] & SER_TX_QUEUED) ) {
          angel_DeviceStatus[DI_SERIAL] |= SER_TX_QUEUED;
#if !defined(MINIMAL_ANGEL) || MINIMAL_ANGEL == 0
          Angel_SerialiseTask( 0,
                               serial_ctrl.tx_processing, 
                               DI_SERIAL,
                               empty_stack,
                               &Angel_GlobalRegBlock[RB_Interrupted]);
#else
          serial_ctrl.tx_processing((void *)DI_SERIAL);
#endif
      }
    } /* IntSource & TX_INTERRUPT */

    else {
        /* Unknown interrupt source */
#if DEBUG_EB285
        puthex(IntSource) ;
        putstring(" Ov") ;
#endif
        /* clear out any errors by reading the line status register */
        IntStatus = GET_STATUS(HOST_COMPORT) ;
    }
}

#pragma check_stack

/* Routines for logging debug information to a seperate serial port */

void logserial_Reset(int port, int baudvalue)
{
    /*
     * Interrupt Enable Register - all interrupts off
     */
    xoutb(0, port + UART_IER) ;

    /*
     * wait for current transmits to finish
     */
    while ( ! LOG_TX_EMPTY(port) )
        /* do nothing */
        ;

    /*
     * Modem Control Register - DTR* and RTS* outputs high; INT
     * pin tri-stated; Rx and Tx "disabled" by setting chip in
     * loopback mode.
     */
        xoutb(0x10, port + UART_MCR) ;

    /*
     * FIFO Control Register - FIFO enabled, but Rx and Tx
     * FIFOs empty; RxRDY and TxRDY mode 1; FIFO interrupt
     * trigger level = 8.
     */
        xoutb(0x8F, port + UART_LCR) ;

    /*
     * Enable divisor latch via the Line Control Register, and
     * set the baud rates
     */
        /* set the baud rate and a word length of 8 */
        xoutb(LCR_DLAB, port + UART_LCR) ;
        xoutb((baudvalue & 0xFF), port + UART_DLLSB) ;
        xoutb(((baudvalue >> 8) & 0xFF), port + UART_DLMSB) ;
        xoutb(0x03, port + UART_LCR) ;

    /*
     * take the chip out of loopback, and enable interrupts; DTR* and
     * RTS* outputs are forced low.
     */
        xoutb(0x0B, port + UART_MCR) ;
}

/* EOF serial.c */
