/****************************************************************************
**                                                                          *
**  Copyright  1997						     								*
**  by Digital Equipment Corporation, Maynard, Massachusetts.		     	*
**                                                                          *
**  All Rights Reserved							     						*
**                                                                          *
**  Permission  is  hereby  granted  to  use, copy, modify and distribute   *
**  this  software  and  its  documentation,  in  both  source  code  and   *
**  object  code  form,  and without fee, for the purpose of distribution   *
**  of this software  or  modifications  of this software within products   *
**  incorporating  an  integrated   circuit  implementing  Digital's  		*
**  StrongARM architecture,  regardless  of the  source of such integrated 	*
**	circuit, provided that the  above copyright  notice and this permission *
**  notice appear  in  all copies,  and  that  the  name  of  Digital  		*
**  Equipment Corporation  not  be  used  in advertising or publicity 		*
**  pertaining to distribution of the  document  or  software without 		*
**  specific, written prior permission.							            *
**                                                                          *
**  Digital  Equipment  Corporation   disclaims  all   warranties  and/or   *
**  guarantees  with  regard  to  this  software,  including  all implied   *
**  warranties of fitness for  a  particular purpose and merchantability,   *
**  and makes  no  representations  regarding  the use of, or the results   *
**  of the use of, the software and documentation in terms of correctness,  *
**  accuracy,  reliability,  currentness  or  otherwise;  and you rely on   *
**  the software, documentation and results solely at your own risk.	    *
**                                                                          *
**  StrongARM is a trademark of Digital Equiptment Corporation              *
**                                                                          *
*****************************************************************************/

#ifndef LINT
extern char *rcsid;
#endif

/*
 * This module provides a common interface to all ethernet devices
 *
 * $Log: edevice.c,v $
 * Revision 2.0  1997/07/23  13:26:04  smithg
 * ebsa285 revisions
 *
 * Revision 1.2  1997/07/23 10:57:10  smithg
 * Header updated
 *
 * Revision 1.1  1997/07/14 14:45:49  smithg
 * Initial revision
 *
 * Revision 1.3  1997/07/04 10:03:34  smithg
 * changed to unix format from dos format
 *
 * Revision 1.2  1997/07/03 10:32:32  smithg
 * system.h includes removed
 *
 * Revision 1.1  1996/03/28 11:14:52  rusling
 * Initial revision
 *
 * Revision 1.2  1996/01/15  10:27:34  rusling
 * First working version.
 *
 * Revision 1.1  1996/01/09  14:26:48  berent
 * Initial revision
 *
 * Revision 1.11  1994/08/05  20:17:17  fdh
 * Updated Copyright header and RCS $Id: identifier.
 *
 * Revision 1.10  1994/06/28  20:11:47  fdh
 * Modified filenames and build precedure to fit into a FAT filesystem.
 *
 * Revision 1.9  1994/06/22  15:11:46  rusling
 * Fixed up WNT compile warnings.
 *
 * Revision 1.8  1994/06/03  20:04:41  fdh
 * Replaced <stdio.h> with <stddef.h>.
 *
 * Revision 1.7  1994/01/19  10:55:52  rusling
 * Ported to Alpha Windows NT.
 *
 * Revision 1.6  1993/11/22  14:57:43  rusling
 * Added support for eshow command, plus some post
 * merge fixups.
 *
 *
 * Revision 1.2  1993/10/01  15:58:43  berent
 * Added module initialisation function
 *
 * Revision 1.1  1993/08/06  09:51:52  berent
 * Initial revision
 *
 *
 */
#include "net_buff.h"
#include "netparam.h"
#include "edevice.h"
#include "debugio.h"
#include "errors.h"
#include "devconf.h"

#if (ETHERNET_SUPPORTED == 1)


/* Device registration structure */
static struct {
    int entry_in_use;
    int (*init_func)(int device_no);
    int (*read_func)(int device_no, 
      unsigned char* balance_buffer, 
      unsigned char* *frame_buffer);
    int (*write_func)(int device_no, 
      unsigned char* frame, 
      int size, 
      unsigned char* frame_buffer, 
      unsigned char* *balance_buffer);
    int (*flush_func)(int device_no);
    void (*stats_func)(int device_no);
    void (*preg_func)(int device_no);
    void (*clear_interrupts_func)(int device_no);
    int (*hw_addr_func)(int device_no, mac_addr hw_address);
    char *description;
} device_table[MAX_ETHER_DEVICES];

/* ether_device_init - initialises an ethernet device
 *
 * Arguments:
 *
 *    device_no    - device number
 *
 * Returned value
 *    0         - initialised succesfully
 *    -ve        - error on initialisation
 */
int ether_device_init(int device_no)
{
    if (device_no >= 0 && 
      		device_no < MAX_ETHER_DEVICES && 
      		device_table[device_no].entry_in_use) {
		return (*device_table[device_no].init_func)(device_no);
    } else {
		return ERR_INITUNKNOWNDEVICE;
    }
}


/* ether_device_read - reads a frame from an ethernet device
 *
 * Arguments:
 *
 *    device_no      - device number
 *    balance_buffer - spare buffer to maintain buffer balance
 *    frame_buffer   - returned buffer containing frame
 *
 * Returned value:
 *
 *    positive     - size of frame read; frame copied into buffer if <= size.
 *    negative     - error on read
 *
 * DESCRIPTION:
 *    This function reads a frame from the device if there is one ready to be read. The
 *    frame read is returned in frame_buffer.  To maintain the driver's pool of buffers
 *    the caller must give the driver an unused buffer (balance_buffer) before the
 *    driver returns the frame.  Once this function has been called balance_buffer must
 *    not be used by the calling software.
 *
 *    On return frame_buffer will always point to a valid buffer.  If the read has failed
 *    (result <= 0) the contents of this buffer will not be valid.  Note that whether or
 *    not the read succeeds it is undefined whether the same buffer will be returned as
 *    frame_buffer as was passed down as balance_buffer.
 */
extern int ether_device_read(int device_no, 
  unsigned char* balance_buffer, 
  unsigned char* *frame_buffer)
{
    if (device_no >= 0 && 
      device_no < MAX_ETHER_DEVICES && 
      device_table[device_no].entry_in_use) {
	return (*device_table[device_no].read_func)(device_no, balance_buffer,
	  frame_buffer);
    } else {
	*frame_buffer = balance_buffer;
	return ERR_READUNKNOWNDEVICE;
    }
}

/* ether_device_write - queue a frame to be sent on an ethernet device
 *
 * Arguments:
 *
 *    device_no         - device number
 *    frame_buffer      - buffer containing frame
 *    frame             - the frame itself
 *    size              - size of frame
 *    balance_buffer    - returned buffer to maintain buffer balance
 *
 * Returned value:
 *
 *    success - len of buf sent
 *    -ve - unable to send
 *
 * DESCRIPTION:
 *    This function writes a frame to the device  To maintain the caller's pool of buffers
 *    the driver must give the caller an unused buffer (balance_buffer) in return for
 *    the buffer containing the frame to be transmitted.  
 *
 *    If the send succeeds then frame_buffer must not be accessed by the caller after this
 *    call.  If the send can't be queued then frame_buffer will remain valid and
 *    the returned value of balance_buffer is undefined and must not be used.
 *
 */
extern int ether_device_write(int device_no, 
  unsigned char* frame, 
  int size, 
  unsigned char* frame_buffer, 
  unsigned char* *balance_buffer)
{
    if (device_no >= 0 && 
      device_no < MAX_ETHER_DEVICES && 
      device_table[device_no].entry_in_use) {
	return (*device_table[device_no].write_func)(device_no, frame, size,
	  frame_buffer, balance_buffer);
    } else {
	return ERR_WRITEUNKNOWNDEVICE;
    }
}

/* ether_device_flush - wait for all writes on an ethernet device to complete
 *
 * Argument:
 *
 *    device_no - device to be flushed, -1 means flush all devices
 */
extern int ether_device_flush(int device_no)
{
    if (device_no >= 0 && 
      device_no < MAX_ETHER_DEVICES && 
      device_table[device_no].entry_in_use) {
	return (*device_table[device_no].flush_func)(device_no);
    } else {
	return ERR_FLUSHUNKNOWNDEVICE;
    }
}

/* ether_device_print_stats - print device statistics
 *
 * Argument:
 *
 *     device_no - device number, -1 means all devices
 */
extern void ether_device_print_stats(int device_no)
{
    if (device_no >= 0 && 
      device_no < MAX_ETHER_DEVICES && 
      device_table[device_no].entry_in_use) {
	(*device_table[device_no].stats_func)(device_no);
    } }
/* ether_device_preg - print device registers
 *
 * Argument:
 *
 *     device_no - device number
 */
void ether_device_preg(int device_no)
{
    if (device_no >= 0 && 
      device_no < MAX_ETHER_DEVICES && 
      device_table[device_no].entry_in_use) {
	(*device_table[device_no].preg_func)(device_no);
    } 
}

/* ether_device_clear_interrupts - clear all interrupts from an ethernet device
 *
 * Argument:
 *
 *     device_no - device number, -1 means all devices
 */
void ether_device_clear_interrupts(int device_no)
{
    if (device_no >= 0 && 
      device_no < MAX_ETHER_DEVICES && 
      device_table[device_no].entry_in_use) {
	(*device_table[device_no].clear_interrupts_func)(device_no);
    } 
}

/* ether_device_get_hw_address - gets the hardware mac address of the device
 *
 * Arguments:
 *
 *     device_no - device_number
 *     hw_address - returned hardware address
 */
int ether_device_get_hw_address(int device_no, mac_addr hw_address)
{
    if (device_no >= 0 && 
      device_no < MAX_ETHER_DEVICES && 
      device_table[device_no].entry_in_use) {
	return (*device_table[device_no].hw_addr_func)(device_no, hw_address);
    } 
return ERR_HWUNKNOWNDEVICE;
}

/* ether_device_next - get the next valid device number
 *
 * Argument:
 *
 *    previous_device_number - previous device number; or -1 if no previous device.
 *
 * Result:
 *
 *    next valid device number or -1 if no more devices.
 *
 * Description:
 *    The purpose of this function it to allow the device table to be scanned.
 *    If it called initially with -1 and then repeatedly called with its previous
 *    return value as its argument it will return each valid device number 
 *    precisely once before returning -1.
 *
 * Notes:
 *    1) The device numbers will not neccesary be returned in assending order
 *    2) If previous_device_number is not the number of a valid device (or -1) the 
 *        result is undefined.
 */
int ether_device_number(int previous_device_number)
{

    /* Loop doing nothing until we find the next entry marked as in use */
    while (++previous_device_number < MAX_ETHER_DEVICES && 
      !device_table[previous_device_number].entry_in_use)
	;

    /* Return either the valid device number we have found; or -1 */
    return (previous_device_number == MAX_ETHER_DEVICES) ? ERR_OUTOFDEVICESPACE :
      previous_device_number;
}


/* ether_device_register - register a new ethernet device
 *
 * Arguments:
 *     device_no              - device_number
 *     init_func              - initialisation function
 *     read_func              - read function
 *     write_func             - write function
 *     flush_func             - flush function
 *     stats_func             - print statistics function
 *     preg_func              - print device registers function
 *     clear_interrupts_func  - clear interrupts function
 *     hw_addr_func           - get hardware address function
 *     description            - description of device.
 */
void ether_device_register(int device_no, 
  int (*init_func)(int device_no), 
  int (*read_func)(int device_no, 
  unsigned char* balance_buffer, 
  unsigned char* *frame_buffer), 
  int (*write_func)(int device_no, 
  unsigned char* frame, 
  int size, 
  unsigned char* frame_buffer, 
  unsigned char* *balance_buffer), 
  int (*flush_func)(int device_no), 
  void (*stats_func)(int device_no), 
  void (*preg_func)(int device_no), 
  void (*clear_interrupts_func)(int device_no), 
  int (*hw_addr_func)(int device_no, mac_addr hw_address), 
  char *description)
{
    if (device_no >= 0 && 
      device_no < MAX_ETHER_DEVICES) {
	if (!device_table[device_no].entry_in_use)
	  {
	    device_table[device_no].entry_in_use = TRUE;
	    device_table[device_no].init_func = init_func;
	    device_table[device_no].read_func = read_func;
	    device_table[device_no].write_func = write_func;
	    device_table[device_no].flush_func = flush_func;
	    device_table[device_no].stats_func = stats_func;
	    device_table[device_no].preg_func = preg_func;
	    device_table[device_no].clear_interrupts_func =
	      clear_interrupts_func;
	    device_table[device_no].hw_addr_func = hw_addr_func;
	    device_table[device_no].description = description;
	}
	 
    } 
}
/* ether_device_init_module - initialise or reinitialise the ether_device module
 */
void ether_device_init_module(void) 
{
    int i;

    for (i = 0; i < MAX_ETHER_DEVICES; i++) {
	device_table[i].entry_in_use = FALSE;
    }
}

/*
 * ether_device_show - show all of the ethernet devices that have been
 * registered.
 */
void ether_device_show(void) 
{
    return; 

  
}


#endif
