/****************************************************************************
**                                                                          *
**  Copyright  1997						     								*
**  by Digital Equipment Corporation, Maynard, Massachusetts.		     	*
**                                                                          *
**  All Rights Reserved							     						*
**                                                                          *
**  Permission  is  hereby  granted  to  use, copy, modify and distribute   *
**  this  software  and  its  documentation,  in  both  source  code  and   *
**  object  code  form,  and without fee, for the purpose of distribution   *
**  of this software  or  modifications  of this software within products   *
**  incorporating  an  integrated   circuit  implementing  Digital's  		*
**  StrongARM architecture,  regardless  of the  source of such integrated 	*
**	circuit, provided that the  above copyright  notice and this permission *
**  notice appear  in  all copies,  and  that  the  name  of  Digital  		*
**  Equipment Corporation  not  be  used  in advertising or publicity 		*
**  pertaining to distribution of the  document  or  software without 		*
**  specific, written prior permission.							            *
**                                                                          *
**  Digital  Equipment  Corporation   disclaims  all   warranties  and/or   *
**  guarantees  with  regard  to  this  software,  including  all implied   *
**  warranties of fitness for  a  particular purpose and merchantability,   *
**  and makes  no  representations  regarding  the use of, or the results   *
**  of the use of, the software and documentation in terms of correctness,  *
**  accuracy,  reliability,  currentness  or  otherwise;  and you rely on   *
**  the software, documentation and results solely at your own risk.	    *
**                                                                          *
**  StrongARM is a trademark of Digital Equipment Corporation               *
**                                                                          *
*****************************************************************************/

#ifndef LINT
extern char *rcsid;
#endif

/*
 *
 */
#include "net_buff.h"
#include "netparam.h"
#include "ethernet.h"
#include "edevice.h"
#include "support.h"
#include "debugio.h"
#include "errors.h"
#include "devconf.h"

#if (ETHERNET_SUPPORTED == 1)


/* Define the structure of an ethernet frame header; this is defined through constants to avoid any possible
 * problems with struct alignment etc.
 */

#define EH_DESTINATION_OFFSET 0
#define EH_SOURCE_OFFSET (EH_DESTINATION_OFFSET+MAC_ADDRESS_SIZE)
#define EH_PROTOCOL_ID_OFFSET (EH_SOURCE_OFFSET+MAC_ADDRESS_SIZE)
#define EH_HEADER_SIZE (EH_PROTOCOL_ID_OFFSET+ETHERNET_PROTOCOL_ID_SIZE)

/* Define the table of registered protocols */
#define MAX_REGISTERED_ETHERNET_PROTOCOLS 5 /* ARP, RARP, IP  */

static struct
  {
    int in_use;
    ethernet_packet_handler processing_function;
    ethernet_protocol_id protocol_id;
    void *protocol_data;
} protocol_table[MAX_ETHER_DEVICES][MAX_REGISTERED_ETHERNET_PROTOCOLS];

/* ethernet_register_protocol - registers an ethernet protocol to be recognised by this module
 *
 * Arguments:
 *
 *    device_no             - device on which the protocol is to be registered
 *    protocol_id           - protocol id to be recognised by this module
 *    processing_function   - function to process received packets; if NULL cancels registration of protocol.
 *    protocol_data         - arbitrary data to be passed back when a packet is received
 *		Returns
 		0 - success
		-ve - error
 */
int ethernet_register_protocol(int device_no, 
  ethernet_protocol_id protocol_id, 
  ethernet_packet_handler processing_function, 
  void *protocol_data)
{

    /* The function scans the table of protocols to see if this protocol is
     * already registered, if it is the entry is updated; if not it puts an
     * entry for the protocol in the first free slot. When a packet is received
     * the table is scanned from begining to end so this should mean that the
     * most commonly used protocols (assumed to be the first ones registered)
     * are found first. */
    int i;
    int first_free_slot = MAX_REGISTERED_ETHERNET_PROTOCOLS;
					/* Used to keep track of whether their 
					 * are any free slots  and if so where;
					 * the value MAX_REGISTERED_PROTOCOLS
					 * is  used to mean that there are no
					 * free slots */
    int current_slot;

    /* Scan from end backwards to simplify setting first_free_slot */
    for (current_slot = MAX_REGISTERED_ETHERNET_PROTOCOLS - 1;
      							current_slot >= 0; current_slot--) {
		if (protocol_table[device_no][current_slot].in_use) {
		    if (__rt_memcmp(protocol_table[device_no][current_slot].protocol_id,
		      		protocol_id, ETHERNET_PROTOCOL_ID_SIZE) == 0) {

				/* Entry already exists; update it */
				if (processing_function == NULL) {

				    /* Remove the entry */
				    protocol_table[device_no][current_slot].in_use = FALSE;
				} 
				else {
				    /* Update the entry */
				    protocol_table[device_no][
				      current_slot].processing_function = processing_function;
				    protocol_table[device_no][current_slot].protocol_data =
				      protocol_data;
				}
			return 0;
		    }
		} 
		else {
		    first_free_slot = current_slot;
		}
    }

    /* If this point is reached a new entry is needed */
    if (first_free_slot != MAX_REGISTERED_ETHERNET_PROTOCOLS) {

		/* There is a slot free; fill it in */
		for (i = 0; i < ETHERNET_PROTOCOL_ID_SIZE; i++) {
		    protocol_table[device_no][first_free_slot].protocol_id[i] =
		      protocol_id[i];
		}
		__rt_memcpy(protocol_table[device_no][first_free_slot].protocol_id,
		  protocol_id, ETHERNET_PROTOCOL_ID_SIZE);
		protocol_table[device_no][first_free_slot].processing_function =
		  processing_function;
		protocol_table[device_no][first_free_slot].protocol_data =
		  protocol_data;
		protocol_table[device_no][first_free_slot].in_use = TRUE;
		return 0;
    }
    else {
		return ERR_EOUTOFPROTOCOLSPACE;
    }
}





/* ethernet_process_one_packet - reads and processes one packet from the ethernet
 *
 * Arguments:
 *
 *    device_no - device to poll
 *    input_buffer - free buffer usable by the ethernet module
 *    output_buffer - returned free buffer that may be used by the caller.
 *
 * Return value:
 *    0 - packet received and processed
 *    -VE - nothing to receive.
 *
 * DESCRIPTION:
 *    This function checks whether there are any packets available to be processed on
 *    the ethernet device. If there are it reads and processes one packet. The caller
 *    must give it a buffer to work with (input_buffer) and will be returned a, possibly
 *    different, buffer (output_buffer) Note that this returned buffer does not contain the 
 *    received packet, or any other meaningfull data; it is returned simply to ensure that caller 
 *    does not run short of buffers.  By calling this function the caller gives up ownership
 *    of input_buffer. The caller must not access input_buffer after calling this function.
 *
 */
int ethernet_process_one_packet(int device_no, 
  unsigned char* input_buffer, 
  unsigned char* *output_buffer)
{
    unsigned char* data_buffer;
    int const count = ether_device_read(device_no, input_buffer, &data_buffer);


    /* Set default for buffer to be returned */
    *output_buffer = data_buffer;

	if (count == 0)
		return ERR_ENOWTRECVD;
		
    if (count < 0) {

	return count;
    }
	
	ethernet_input(device_no, data_buffer, count, data_buffer, output_buffer);
	/* Packet read, return TRUE whether it was passed on the next layer or 
	 * not since this  tells the caller that there may be more packets */
	 /* Don't want the balance buffer */
	 
	return 0;
}

/* ethernet_input - processes the packet passed to it as an ethernet frame
 *
 * Arguments:
 *
 *    device_no - device on which frame was received
 *    frame - the frame to be processed
 *    size - size of frame
 *    input_buffer - the buffer containing the frame
 *    balance_buffer - returned free buffer that may be used by the caller.
 *
 */
void ethernet_input(int device_no, 
  unsigned char* frame, 
  int size, 
  unsigned char* input_buffer, 
  unsigned char* *balance_buffer)
{

    /* Search for the protocol id in the table */
    int i;
    mac_addr source;
    int current_entry;
    for (current_entry = 0; current_entry < MAX_REGISTERED_ETHERNET_PROTOCOLS;
      current_entry++) {
		if (protocol_table[device_no][current_entry].in_use && 
		  __rt_memcmp(protocol_table[device_no][current_entry].protocol_id, 
		  frame + EH_PROTOCOL_ID_OFFSET, 
		  ETHERNET_PROTOCOL_ID_SIZE) == 0) {

		    /* Copy the source address out of the buffer to avoid problems when
		     * the buffer is reused */
		    for (i = 0; i < MAC_ADDRESS_SIZE; i++)
				source[i] = *(unsigned char *)(frame + EH_SOURCE_OFFSET + i);
		    /* Pass on the packet by calling the protocol's processing function
		     */
		    (*(protocol_table[device_no][current_entry].processing_function))
		      (protocol_table[device_no][current_entry].protocol_data, 
		      device_no, 
		      frame + EH_HEADER_SIZE, 
		      size - EH_HEADER_SIZE, 
		      source, 
		      input_buffer, 
		      balance_buffer);

		    /* And stop scanning the table since there can only be one entry
		     * per protocol id */
		    break;
		}
    }
}

/* ethernet_write - sends a packet on an ethernet
 *
 * Arguments:
 *
 *    device_no - device on which to send the packet
 *    packet    - pointer to data packet to be sent
 *    destination - destination address for packet
 *    protocol_id - protocol id for packet
 *    size        - size of packet
 *    frame_buffer - buffer containing packet
 *    balance_buffer - buffer returned to maintain buffer balance
 *
 * Return values:
 *
 *    0 - send successfull
 *    -VE - Error occured
 *
 * DESCRIPTION:
 *    This function writes an ethernet data packet to the device  To maintain the caller's 
 *    pool of buffers the driver must give the caller an unused buffer (balance_buffer) 
 *    in return for the buffer containing the frame to be transmitted.  
 *
 *    If the send succeeds then frame_buffer must not be accessed by the caller after this
 *    call.  If the send can't be queued then frame_buffer will remain valid and
 *    the returned value of balance_buffer is undefined and must not be used.
 *
 */
int ethernet_write(int device_no, 
  unsigned char* packet, 
  mac_addr destination, 
  ethernet_protocol_id protocol_id, 
  int size, 
  unsigned char* frame_buffer, 
  unsigned char* *balance_buffer)
{
    mac_addr source;
    int i;

    /* Work out where the start of the ethernet header should be */
    unsigned char* const header_start = packet - EH_HEADER_SIZE;
    
	
	/* Check that there is enough space in the buffer for the header */
	*balance_buffer = frame_buffer;
    if (header_start < frame_buffer) {
		return ERR_EBUFFERTOOSMALL;
    }

    /* Get the source address */
    ether_device_get_hw_address(device_no, source);

    /* Set up the header */
    for (i = 0; i < MAC_ADDRESS_SIZE; i++)
		*(unsigned char *)(header_start + EH_SOURCE_OFFSET + i)= source[i];
    for (i = 0; i < MAC_ADDRESS_SIZE; i++)
		*(unsigned char *)(header_start + EH_DESTINATION_OFFSET + i)=
	  		destination[i];
    for (i = 0; i < ETHERNET_PROTOCOL_ID_SIZE; i++)
		*(unsigned char *)(header_start + EH_PROTOCOL_ID_OFFSET + i)=
	  		protocol_id[i];
    /* And send it */
    return ether_device_write(device_no, header_start, size + EH_HEADER_SIZE,
      frame_buffer, balance_buffer);
}

/* ethernet_data_offset - returns the offset of the data in an ethernet frame
 *
 * Return value:
 *    Offset.
 */
int ethernet_data_offset(void) 
{
    return EH_HEADER_SIZE;
}
/* ethernet_printpacket - print the contents of an ethernet frame
 *	NOT USED BY ANGEL - removed
 * Arguments:
 *    p          - the frame
 *    frame_size - its size
 */

void ethernet_printpacket(unsigned char* p, unsigned short frame_size)
{
	p=p; frame_size=0;
}



/* ethernet_printaddress - utility to print an ethernet address 
 *		NOT USED  BY ANGEL so removed
 * Argument:
 *    address - MAC address to be printed
 */
void ethernet_printaddress(mac_addr address)
{
return;	
}

#endif




