/****************************************************************************
**                                                                          *
**  Copyright  1997						     								*
**  by Digital Equipment Corporation, Maynard, Massachusetts.		     	*
**                                                                          *
**  All Rights Reserved							     						*
**                                                                          *
**  Permission  is  hereby  granted  to  use, copy, modify and distribute   *
**  this  software  and  its  documentation,  in  both  source  code  and   *
**  object  code  form,  and without fee, for the purpose of distribution   *
**  of this software  or  modifications  of this software within products   *
**  incorporating  an  integrated   circuit  implementing  Digital's  		*
**  StrongARM architecture,  regardless  of the  source of such integrated 	*
**	circuit, provided that the  above copyright  notice and this permission *
**  notice appear  in  all copies,  and  that  the  name  of  Digital  		*
**  Equipment Corporation  not  be  used  in advertising or publicity 		*
**  pertaining to distribution of the  document  or  software without 		*
**  specific, written prior permission.							            *
**                                                                          *
**  Digital  Equipment  Corporation   disclaims  all   warranties  and/or   *
**  guarantees  with  regard  to  this  software,  including  all implied   *
**  warranties of fitness for  a  particular purpose and merchantability,   *
**  and makes  no  representations  regarding  the use of, or the results   *
**  of the use of, the software and documentation in terms of correctness,  *
**  accuracy,  reliability,  currentness  or  otherwise;  and you rely on   *
**  the software, documentation and results solely at your own risk.	    *
**                                                                          *
**  StrongARM is a trademark of Digital Equipment Corporation               *
**                                                                          *
*****************************************************************************/

#ifndef LINT
extern char *rcsid;
#endif

/*
 * This module provides a limited end system implementation of the IP protocol. 
 * In particular it does not support the receipt of fragmented datagrams; and 
 * ignores all IP header options.
 * 
 */
#include "net_buff.h"
#include "netparam.h"
#include "ether.h"
#include "address.h"
#include "ip.h"
#include "ethernet.h"
#include "support.h"
#include "debugio.h"
#include "arm_ether.h"
#include "errors.h"

#include "devconf.h"

#if (ETHERNET_SUPPORTED == 1)
/* 
 * IP header structure
 */
#define IPH_VER_IHL_OFFSET   0                                     /* Version and header size */
#define IPH_SERVICE_OFFSET   (IPH_VER_IHL_OFFSET+1)                /* Service type */
#define IPH_LENGTH_OFFSET    (IPH_SERVICE_OFFSET+1)                /* Total packet length */
#define IPH_ID_OFFSET        (IPH_LENGTH_OFFSET+2)                  /* ID for fragmentation */
#define IPH_FRAG_FLAG_OFFSET (IPH_ID_OFFSET+2)                     /* Fragmentation control */
#define IPH_TTL_OFFSET       (IPH_FRAG_FLAG_OFFSET+2)              /* Time to live */
#define IPH_PROTOCOL_OFFSET  (IPH_TTL_OFFSET+1)                    /* Upper layer protocol id */
#define IPH_CHKSUM_OFFSET    (IPH_PROTOCOL_OFFSET+1)               /* Header checksum */
#define IPH_SOURCE_OFFSET    (IPH_CHKSUM_OFFSET+2)                 /* Source address */
#define IPH_DEST_OFFSET      (IPH_SOURCE_OFFSET+IP_ADDRESS_SIZE)   /* Destination address */
#define IPH_OPTIONS_OFFSET   (IPH_DEST_OFFSET+IP_ADDRESS_SIZE)     /* Start of options, if any */


/* Define the table of registered protocols */
#define MAX_REGISTERED_IP_PROTOCOLS 3 /* UDP */

typedef struct
{
    ip_addr ip_address; /* IP address of device */
    struct { 
	int in_use;
	ip_packet_handler processing_function;
	ip_protocol_id protocol_id;
	void * protocol_data;
    } protocol_table[MAX_REGISTERED_IP_PROTOCOLS];
} device_entry;

static device_entry device_table[MAX_ETHER_DEVICES];


/*
 * Internal functions
 */


/* ip_input - ip packet processing function. Read and process a received packet
 *
 * Arguments:
 *
 *     dt_entry      - device table entry for device on which packet was received;
 *     device_no     - device on which packet was received;
 *     packet        - packet, with ethernet header removed;
 *     size          - size of packet;
 *     source        - source MAC address of packet;
 *     frame_buffer  - original buffer containing packet.
 *     balance_buffer- buffer returned to maintain buffer balance buffer.
 *
 * Description:
 *     Processes a received IP packet. For a detailed explanation of the 
 *     interface see ethernet.h.
 *
 * Note:
 *     This function is declared static to prevent other modules calling it directly (rather
 *     than through function pointers.).
 */
static void ip_input(void * dt_entry,
		     int device_no,
		     unsigned char* packet,
		     int size,
		     mac_addr source,
		     unsigned char* frame_buffer,
		     unsigned char* * balance_buffer)
{
    device_entry * const device_table_entry = (device_entry * const) dt_entry;
    int const header_size = *(unsigned char *)(packet+IPH_VER_IHL_OFFSET) & 0x0f; /* In 32 bit units */
    int const packet_size = (*(unsigned char *)(packet+IPH_LENGTH_OFFSET) << 8) 
								+ *(unsigned char *)(packet+IPH_LENGTH_OFFSET+1); /* In bytes */
    int header_index;
    unsigned int header_sum = 0;
    int current_entry;
    ip_addr ip_source;
    ip_addr ip_dest;
    int i;


   	/* Set up the default buffer to be returned */
    *balance_buffer = frame_buffer;
    /*
     * Start by checking that the header is valid
     */
    if ( (*(unsigned char *)(packet+IPH_VER_IHL_OFFSET) >> 4) != 4) {/* IP protocol version 4 */  

		return;				 	
	}
    if (header_size < 5) {	/* Header length at least 5 * 32 bits */

		return;
	}
    if ( (*(unsigned char *)(packet+IPH_FRAG_FLAG_OFFSET) & 0x3f) != 0) { /* Not fragmented (last fragment 
										     & most significant part of offset 0) */
		return;
	}
		
    if ( *(unsigned char *)(packet+IPH_FRAG_FLAG_OFFSET+1) != 0) { /* Not fragmented (least significant part of offset 0) */
		return;
	}
    if (*(unsigned char *)(packet+IPH_TTL_OFFSET) == 0) {/* Time to live > 0 */
		return;
	}
    if (packet_size > size)	{/* packet fits in frame */

		return;
	}
		
    if (packet_size < header_size * 4) { /* packet longer than header */	
		return;
	}
       
    /* 
     * Check the destination address; if the device's address is not yet known accept anything. If
     * it is accept only the device's address and network and subnet broadcasts. Acceptable 
     * broadcast addresses start with bytes matching this device's address; and end with bytes 
     * containing 0xff (see rfc1340; this implementation will actually accept some addresses 
     * that are not valid broadcast addresses)
	 */
    if(__rt_memcmp(device_table_entry -> ip_address, own_ip_addr_unknown_address, IP_ADDRESS_SIZE) != 0) {
		int address_index;
		for(address_index = 0; address_index < IP_ADDRESS_SIZE; address_index++)
		    if(*(unsigned char *)(packet+IPH_DEST_OFFSET + address_index) != device_table_entry -> ip_address[address_index])
			break;

		/* Now check the remaining bytes (if any) for equality to 0xff; if the address has already
	           been matched this loop will never execute */

		for(;address_index < IP_ADDRESS_SIZE ; address_index++)
		    if(*(unsigned char *)(packet+IPH_DEST_OFFSET + address_index) != 0xff) {
			/* Not device or broadcast address; ignore packet */
			return;
		    }
    }
	
    /* 
     * Check the checksum; the ones complement sum of the 16 bit header words should be 0xffff; 
     * To calculate the ones complement sum calculate the 32 bit sum and add the top and
     * bottom words together.
     */
    for(header_index = 0; header_index < header_size * 4; header_index += 2) {
	header_sum += (*(unsigned char *)(packet+header_index) << 8) + *(unsigned char *)(packet+header_index+1);
    }
    header_sum = (header_sum & 0xffff) + ((header_sum >> 16) & 0xffff);
    header_sum += (header_sum >> 16);
    
    if(header_sum != 0xffff) {
		/* Bad packet header checksum */
		return;
    }

    /*
     * Packet acceptable. Process it.
     */

    /* Search for the protocol id in the table */
    for(current_entry = 0; current_entry < MAX_REGISTERED_IP_PROTOCOLS; current_entry++) {
		if(device_table_entry -> protocol_table[current_entry].in_use &&
		   device_table_entry -> protocol_table[current_entry].protocol_id == *(unsigned char *)(packet+IPH_PROTOCOL_OFFSET)) {
		    /* Copy the addresses to avoid problems if the packet is reused */
			for(i=0; i<IP_ADDRESS_SIZE; i++) {
				ip_source[i] = *(unsigned char *)(packet+IPH_SOURCE_OFFSET+i);
			}


		    for(i=0; i<IP_ADDRESS_SIZE; i++) {
				ip_dest[i] = *(unsigned char *)(packet+IPH_DEST_OFFSET+i);
			}

			/* Just before passing it on store the mac/ip addr combination for ARP */
			arp_add(device_no, ip_source, source);
		    /* Pass on the packet by calling the protocol's processing function */
		    (*(device_table_entry -> protocol_table[current_entry].processing_function))
			(device_table_entry -> protocol_table[current_entry].protocol_data,
			 device_no,
			 packet + header_size * 4,
			 packet_size - header_size * 4,
			 ip_source,
			 ip_dest,
			 frame_buffer,
			 balance_buffer);

		    /* And stop scanning the table since there can only be one entry per protocol id */
		    break;
		}
    }
}
  
/*
 * Externally callable functions
 */

void ip_init(int device_no)
{
    int entry_no;
    
    /* Clear out all table entries for the device  */
    __rt_memcpy(device_table[device_no].ip_address, client_ipaddr, IP_ADDRESS_SIZE);
    for(entry_no = 0; entry_no < MAX_REGISTERED_IP_PROTOCOLS; entry_no++) {
	device_table[device_no].protocol_table[entry_no].in_use = FALSE;
    }

    /* Register the IP protocol with the ethernet protocol handler */
    ethernet_register_protocol(device_no, 
			       ether_protocol_ip, 
			       ip_input,
			       &device_table[device_no]);
}
	

/* ip_set_device_addr - tells ip a device's ip address.
 *
 * Arguments:
 *     device_no - device number
 *     ip        - ip address
 */
void ip_set_device_addr(int device_no,ip_addr ip)
{
    /* Remember the address */
    __rt_memcpy(device_table[device_no].ip_address, ip, IP_ADDRESS_SIZE);
    __rt_memcpy(client_ipaddr, ip, IP_ADDRESS_SIZE);	
    /* And tell the arp module */
    arp_set_device_addr(device_no, ip);
}

/* ip_register_protocol - registers an IP protocol to be recognised in received frames by this module
 *
 * Arguments:
 *
 *    protocol_id           - protocol id to be recognised by this module
 *    processing_function   - function to process received packets; if NULL cancels registration of protocol.
 *    protocol_data         - arbitrary data to be passed back when a packet is received
 *
 * Returned value:
 *    0   - protocol registered
 *    -VE  - unable to register protocol
 */
int ip_register_protocol(int device_no,
			 ip_protocol_id protocol_id,
			 ip_packet_handler processing_function,
			 void * protocol_data)

{
    /* The function scans the table of protocols to see if this protocol is already registered,
     * if it is the entry is updated; if not it puts an entry for the protocol in the first free
     * slot. When a packet is received the table is scanned from begining to end so this should
     * mean that the most commonly used protocols (assumed to be the first ones registered) are
     * found first.
     */
    int first_free_slot = MAX_REGISTERED_IP_PROTOCOLS; /* Used to keep track of whether their are any free slots
								 and if so where; the value MAX_REGISTERED_PROTOCOLS is
								 used to mean that there are no free slots */
    int current_slot;

    /* Scan from end backwards to simplify setting first_free_slot */
    for(current_slot = MAX_REGISTERED_IP_PROTOCOLS-1; current_slot >= 0; current_slot--) {
		if(device_table[device_no].protocol_table[current_slot].in_use) {
		    if(device_table[device_no].protocol_table[current_slot].protocol_id == protocol_id) {
				/* Entry already exists; update it */
				if(processing_function == NULL) {
				    /* Remove the entry */
				    device_table[device_no].protocol_table[current_slot].in_use = FALSE;
				} else {
				    /* Update the entry */
				    device_table[device_no].protocol_table[current_slot].processing_function = processing_function;
				    device_table[device_no].protocol_table[current_slot].protocol_data = protocol_data;
				}
				return 0;
		    }
		} 
		else {
		    first_free_slot = current_slot;
		}
    }
    /* If this point is reached a new entry is needed */
    if(first_free_slot != MAX_REGISTERED_IP_PROTOCOLS) {
		/* There is a slot free; fill it in */
		device_table[device_no].protocol_table[first_free_slot].protocol_id = protocol_id;
		device_table[device_no].protocol_table[first_free_slot].processing_function = processing_function;
		device_table[device_no].protocol_table[first_free_slot].protocol_data = protocol_data;
		device_table[device_no].protocol_table[first_free_slot].in_use = TRUE;
		return 0;
    }
    else {
		return ERR_IPOUTOFPROTOCOLSPACE;
    }
return -1; /* Shouldn't ever get here, shuts up the compiler warnings though */	
}
/* ip_write - sends a packet on an IP network
 *
 * Arguments:
 *
 *    device_no - device on which to send the packet
 *    packet    - pointer to data packet to be sent
 *    destination - destination address for packet
 *    protocol_id - protocol id for packet
 *    size        - size of packet
 *    frame_buffer - buffer containing packet
 *    balance_buffer - buffer returned to maintain buffer balance
 *
 * Return values:
 *
 *    0  - send successful
 *    -VE - Error occured
 *
 * DESCRIPTION:
 *    This function writes an ip data packet to the device  To maintain the caller's 
 *    pool of buffers the driver must give the caller an unused buffer (balance_buffer) 
 *    in return for the buffer containing the frame to be transmitted.  
 *
 *    If the send succeeds then frame_buffer must not be accessed by the caller after this
 *    call.  If the send can't be queued then frame_buffer will remain valid and
 *    the returned value of balance_buffer is undefined and must not be used.
 *
 */
int ip_write(int device_no, 
	     unsigned char*  packet, 
	     ip_addr destination,
	     ip_protocol_id protocol_id,
	     int size, 
	     unsigned char* frame_buffer,
	     unsigned char* * balance_buffer)
{
    /* Work out where to put the IP header; we are not going to put any options in the packet */
    unsigned char*  const header = packet - IPH_OPTIONS_OFFSET;
    int const ip_packet_size = size + IPH_OPTIONS_OFFSET;
    static short int packet_id = 0; /* Used to give each packet a unique id */
    int header_index;
    unsigned int header_sum = 0;
    mac_addr destination_mac_addr;
    int i;


	*balance_buffer = frame_buffer; /* in the error cases */

    /* Check that we have enough space in the buffer for the header */
    if(header - frame_buffer <  ethernet_data_offset()) {
		return ERR_IPBUFFERTOOSMALL;
    }

    /* fill in the header */
    *(unsigned char *)(header+IPH_VER_IHL_OFFSET)= (4 << 4) + 5; /* Version 4 and header size 5 (no options) */
    *(unsigned char *)(header+IPH_SERVICE_OFFSET)= 0; /* Precedence: Routine, Normal Delay, Normal Throughput, Normal Reliablility */
    *(unsigned char *)(header+IPH_LENGTH_OFFSET)= ip_packet_size >> 8; /* MS byte of packet size */
    *(unsigned char *)(header+IPH_LENGTH_OFFSET+1)= ip_packet_size & 0xff; /* LS byte of packet size */
    *(unsigned char *)(header+IPH_ID_OFFSET)= packet_id >> 8; /* MS byte of packet id */
    *(unsigned char *)(header+IPH_ID_OFFSET+1)= (packet_id++) & 0xff; /* LS byte; increment id so that it is different next time */
    *(unsigned char *)(header+IPH_FRAG_FLAG_OFFSET)= 0x40; /* Not fragmented but allow fragmentation */
    *(unsigned char *)(header+IPH_FRAG_FLAG_OFFSET+1)= 0;
    *(unsigned char *)(header+IPH_TTL_OFFSET)= 255; /* This a new packet so give it the maximum time to live */
    *(unsigned char *)(header+IPH_PROTOCOL_OFFSET)= protocol_id; /* Protocol id from argument */
    *(unsigned char *)(header+IPH_CHKSUM_OFFSET)= 0; /* Checksum to 0 for now to make calculation easier later */
    *(unsigned char *)(header+IPH_CHKSUM_OFFSET+1)= 0;

    /* The destination address comes from the arguments */
    for(i=0; i<IP_ADDRESS_SIZE; i++) {
		*(unsigned char *)(header+IPH_DEST_OFFSET+i)=*(unsigned char *)(destination+i);
	}

    /* The source address is the device's IP address */
    for(i=0; i<IP_ADDRESS_SIZE; i++) 
		*(unsigned char *)(header+IPH_SOURCE_OFFSET+i)=*(unsigned char *)(device_table[device_no].ip_address+i);
    
    /* Header complete; work out the checksum. This is "the 16 bit one's complement of the one's
     * complement of all 16 bit words in the header" (rfc 791).
     *
     * To calculate the one's complement sum calculate the 32 bit unsigned sum and add the top and
     * bottom words together.
     */

    for(header_index = 0; header_index < IPH_OPTIONS_OFFSET; header_index += 2) {
		header_sum += (*(unsigned char *)(header+header_index) << 8) + *(unsigned char *)(header+header_index+1);
    }
    header_sum = (header_sum & 0xffff) + (header_sum >> 16);

    if(header_sum == 0xffff)header_sum = 0;

    /* Now take its one's complement and put it in the packet */
    header_sum = ~header_sum;
    *(unsigned char *)(header+IPH_CHKSUM_OFFSET)= header_sum >> 8;
    *(unsigned char *)(header+IPH_CHKSUM_OFFSET+1)= header_sum & 0xff;


    /* Packet complete; use arp to work out the MAC address to send it to */
    if(arp_resolve(device_no, destination, destination_mac_addr)) {

		return ERR_IPARPNOTRESOLVED;
    }




    /* And send it */
    return ethernet_write(device_no,
			  header,
			  destination_mac_addr,
			  ether_protocol_ip,
			  ip_packet_size,
			  frame_buffer,
			  balance_buffer);
}
/* ip_data_offset - returns the offset at which ip data should be placed in a new ethernet frame
 *
 * Return value:
 *    Offset.
 */
int ip_data_offset(void)
{
    return ethernet_data_offset()+IPH_OPTIONS_OFFSET;
}
/* ip_get_address - gets the address that will be used as the source address of transmitted packets.
 *
 * Arguments:
 *     device_no - device number of device on which packet is to be sent.
 *     ip        - returned ip address
 *
 * Note:
 *     This function exists so that UDP can calculate header checksums.
 */
void ip_get_device_address(int device_no, ip_addr ip)
{
    __rt_memcpy(ip,device_table[device_no].ip_address, IP_ADDRESS_SIZE);
}
/* ip_printaddress - utility to print an ethernet address 
 *
 * Argument:
 *    address - IP address to be printed
 */
void ip_printaddress(mac_addr address)
{
  // printf("%d.%d.%d.%d",address[0], address[1],address[2],address[3]);
return; 
  
}
    
#endif

