/****************************************************************************
**                                                                          *
**  Copyright  1997						     								*
**  by Digital Equipment Corporation, Maynard, Massachusetts.		     	*
**                                                                          *
**  All Rights Reserved							     						*
**                                                                          *
**  Permission  is  hereby  granted  to  use, copy, modify and distribute   *
**  this  software  and  its  documentation,  in  both  source  code  and   *
**  object  code  form,  and without fee, for the purpose of distribution   *
**  of this software  or  modifications  of this software within products   *
**  incorporating  an  integrated   circuit  implementing  Digital's  		*
**  StrongARM architecture,  regardless  of the  source of such integrated 	*
**	circuit, provided that the  above copyright  notice and this permission *
**  notice appear  in  all copies,  and  that  the  name  of  Digital  		*
**  Equipment Corporation  not  be  used  in advertising or publicity 		*
**  pertaining to distribution of the  document  or  software without 		*
**  specific, written prior permission.							            *
**                                                                          *
**  Digital  Equipment  Corporation   disclaims  all   warranties  and/or   *
**  guarantees  with  regard  to  this  software,  including  all implied   *
**  warranties of fitness for  a  particular purpose and merchantability,   *
**  and makes  no  representations  regarding  the use of, or the results   *
**  of the use of, the software and documentation in terms of correctness,  *
**  accuracy,  reliability,  currentness  or  otherwise;  and you rely on   *
**  the software, documentation and results solely at your own risk.	    *
**                                                                          *
**  StrongARM is a trademark of Digital Equipment Corporation               *
**                                                                          *
*****************************************************************************/

#include "devconf.h"
#if (ETHERNET_SUPPORTED == 1)

#include "debugio.h"
#include "socket.h"
#include "support.h"
#include "net_buff.h"
#include "ethernet.h"
#include "edevice.h"
#include "errors.h"
#include "arp.h"



/* List of familiy/socket type pairs */
static struct { 	
	unsigned short domain;					/* family name */
	unsigned short supported_types;		/* bitmap of supported types ie. bit 0 = SOCK_STREAM */
	unsigned short def_port[NUM_OF_SOCK_TYPES]; /* Default port num for each domain & type pair / NULL for unsupported */		
} fpp[NUM_SUPPORTED_DOMAINS];


static packet_queue *pq = NULL;
static inuse_socket_table *pIST = NULL;


#define MAX_PQ_AGE 50

/*      		_oOo_              */

/*
	Initialisation
	Function int init_sockets(void)
	Desc:
		Allocs buffers and set default values
*/
void socket_init(void)
{
	fpp[0].domain = AF_INET;
	fpp[0].supported_types = (1 << SOCK_DGRAM);
	fpp[0].def_port[SOCK_DGRAM-1] = (unsigned short)IP_UDP;
	
	socket_buffer_init();

return ;
}



/*
	Function called by underlying layers (UDP) to queue a packet read from the ethernet
*/
static void sock_udp_input(void * protocol_data,
		       int device_no,			/* id of physical hw device */
		       unsigned char* packet,	/* ptr to start of data for us */
		       int size,				/* size of data at packet */
		       udp_port source_port,	/* id of port given at creation */
			   udp_port dest_port,		/* port it is heading for */
		       ip_addr source_address,	/* source IP address */
		       unsigned char* frame_buffer, /* ptr to start of entire packet buffer */
		       unsigned char* * balance_buffer)	/* buffer to be swapped for frame buffer (ETHER_BUFFER_SIZE) */
{
	packet_queue *local_pq,*copy_pq;

	if (NULL == (local_pq = pqmalloc(sizeof(packet_queue))) ) {
		*balance_buffer = frame_buffer; /* no room packet q entry */
		return;
	}
	

	/* hand back balance buffer */
	if (NULL == (*balance_buffer = framemalloc(ETHER_BUFFER_SIZE))) {
		/* No memory, ignore packet */
		pqfree(local_pq);
		*balance_buffer = frame_buffer;
		return;	
	}
	/* Update pq descriptor */

	local_pq->device_no = device_no;
	local_pq->packet = packet;
	local_pq->size = size;
	local_pq->age = 0;
	__rt_memcpy(local_pq->source_port,source_port,UDP_PORT_SIZE);
	__rt_memcpy(local_pq->dest_port,dest_port,UDP_PORT_SIZE);
	__rt_memcpy(local_pq->source_address,source_address,sizeof(ip_addr));
	__rt_memcpy(local_pq->mac, frame_buffer+6,MAC_ADDRESS_SIZE);
	local_pq->frame_buffer = frame_buffer;
	local_pq->next = NULL;
	
	/* Queue frame_buffer and packet for next poll */
	/* Walk list and insert new descriptor */
	if (pq == NULL)
		pq = local_pq;
	else {
		
		for(copy_pq = pq;;copy_pq = copy_pq->next)	
			if (copy_pq->next == NULL) {
				copy_pq->next = local_pq;
				break;
			}
	}
	/* Queued - that's all folks...... */
}



/*
	Asks ethernet layer to check for packets
	This goes against conventions I suppose, asking the top layer to 
	check the bottom layer direct, but it is simple- ie no needs for interrupts.
	Perhaps this is the wrong place for the function to live.

*/
int poke_bottom_layer(void)
{
unsigned char *buf,*bb;
int rtn;


	if (NULL == (buf = framemalloc(ETHER_BUFFER_SIZE))) {
		return FALSE;
	}

	/* Returns TRUE if it finds a packet. NOTE: the buffer here does
		not contain the data. */
	rtn = ethernet_process_one_packet(DEVICE,buf,&bb);
	framefree(bb);
 	
return rtn;
}

/*
	function: find_socket(int socket)
	Desc: Find table entry with socket labelled 'socket'
	
	return:
		NULL on failure
		table pointer on success
*/
static inuse_socket_table *find_socket(int socket)
{
	inuse_socket_table *ist;
	
	ist = pIST; /* point to 1st entry or NULL if none */
	while (ist != NULL) {
		if (ist->socket_num == (short)socket)
			return ist;
		ist = ist->next;
	}
	return NULL;
}

/*
	Function: int get_default_port(int af, int type, int *protocol)
	Parameters:
		af - domain
		type - socket type
		protocol - pointer to assigned protocol
	Description:
		Find default protocol entry for given af/type combinaton
		Assigns this value to protocol
		if non found returns error
	returns:
		 -ve on error
		 0 success
	
*/

static int check_for_protocol(short af,short type, udp_port protocol)
{
	int i;
	
	for (i=0;i<NUM_SUPPORTED_DOMAINS;i++)  {
		if ((af == fpp[i].domain) &&   /* Correct domain? */
			(fpp[i].supported_types & (1 << type))) { /* Supported type? */
			return (0); /* Here we should check the supported protocol list, there isn't one! - do later */
		}
	}	
	return ERR_SOCKPROTONOTFOUND; /* Domain not found */	

}
/*
	int bind(int socket, sockaddr_in *localadddr, int addrlen)
	Params:
		socket - previously opened socket
		localaddr- address to bind socket to (eg. internet address)
		addrlen - length of the address
	Description:	
		Binds a socket to a port number		 
	Returns
		0 on success
		-ve - fail
	Restrictions:
		Works only with UDP.
*/

int	bind(int socket, struct sockaddr_in *localaddr , int addrlen)
	{
	int err;
	int zero = 0;
	udp_port port; /* This type is a uchar array by the way */
	unsigned short  af = localaddr->sin_family;		/* address family, AF_INET only */
	unsigned short type = SOCK_DGRAM;	/* SOCK_DGRAM only, should lookup in table */
	inuse_socket_table *ist = find_socket(socket);
	
	if(NULL == ist) { /* No table entry for socket */
		return ERR_SOCKSOCKNOTFOUND;
	}
	
	__rt_memcpy(port,&localaddr->sin_port,UDP_PORT_SIZE); /* copy in network byte order */
	

	/* 	UDP only folks. Find a port or
	check protocol (ie UDP in only at the moment) is valid for af/type combination */
	if (0 != __rt_memcmp(port,&zero,UDP_PORT_SIZE)) {
		if(0 == (err = check_for_protocol(af,type,port))) {
			err = udp_register_well_known_port(DEVICE,port,sock_udp_input,NULL);
		}
	}
	else {
		err = udp_create_port(DEVICE,sock_udp_input,NULL,port);
	}
		
	if (err) {
		return err;	
	}
	
	/* copy  IP address, port, family */
	__rt_memcpy(&localaddr->sin_port,port,UDP_PORT_SIZE); /* assign port number */
	__rt_memcpy(&ist->s.si, localaddr, addrlen);
	__rt_memcpy(ist->port_num, port, UDP_PORT_SIZE);

	return 0; /* success */	
}

/*
	Function: int getsockname (int socket,struct sockaddr *la, int *addrlen)
	Desc:
		find socket and return portnumber bound to it 
	Returns:
		0 : success
		-ve : fail
*/

int     getsockname(int socket, struct sockaddr *localaddr, int *addrlen)
{
	inuse_socket_table *ist;

	ist = find_socket(socket);
	if (NULL == ist) {
		return ERR_SOCKNONAME;
	}
	
	__rt_memcpy(localaddr,&ist->s.sa,sizeof(sockaddr));
	*addrlen = sizeof(sockaddr);
	
	return 0;
}


static packet_queue* remove_packet(packet_queue *find_pq)
{
	packet_queue *lpq, *last_pq;
	
	lpq = pq; 
	last_pq = NULL;
	
  	
	while (lpq != NULL) {
		if (lpq == find_pq) { 
			framefree(lpq->frame_buffer);
			
			if (NULL == last_pq) { /* was 1st entry in Q */
				pq = lpq->next;
				pqfree(lpq);
				lpq = pq;
			}
			else {
				last_pq->next = lpq->next;
				pqfree(lpq);
				lpq = last_pq->next;
			}
			return lpq;				
		}
		else {
			last_pq = lpq;
			lpq = lpq->next;
		}	
	}
	return find_pq;	
}


/*
	See recvfrom,
	recv doesn't return the source of the packet.
*/
int 	recv(int socket, void *buf, int len, int flags)
{
	sockaddr sa;
	int addrlen;
	

return	recvfrom(socket, buf, len, flags, &sa, &addrlen);
}

/*
	Function recvfrom(int socket, void *buf, int len, int flags, struct sockaddr *fromaddr, int *addrlen)
		socket - socket number
		buf - area to write data of max length len
		flags - unused
		fromaddr - empty struct
		addrlen - returned size of struct
	Description:
		Finds socket then checks incoming packet queue for a packet waiting with the appropriate port number,
		which is assigned when the socket is bound. If found the packet is copied, up to max length of len,
		into buf, the sockaddr struct is copied onto fromaddr, and its size into addrlen.
		The packet is then removed from the queue.
	Returns:
		failure: -ve to find packet/error
		success: packet len.
	Notes: 	The packets are not aged, so could hog memory for a long time if not read. 
			There is no non-blocking concept here.
				
*/
int 	recvfrom(int socket, void *buf, int len, int flags, struct sockaddr *fromaddr, int *addrlen)
{
	packet_queue *lpq = pq, *last_pq = NULL;  /* local pq */
	inuse_socket_table *list = pIST; /* local ist */
	int packet_len, j;
	sockaddr_in *fa = (sockaddr_in *)fromaddr;
	ip_addr ipaddr;
	


	if (NULL == (list = find_socket(socket)))
		return ERR_SOCKSOCKNOTFOUND; /* socket not found */
	
	poke_bottom_layer(); /* Check lower layers */
	
	while (NULL != lpq) {
		if ((0 == (__rt_memcmp(&list->s.si.sin_port,lpq->dest_port,UDP_PORT_SIZE)))) 
	   			goto rf_got_port;
		lpq->age++;	
		if (lpq->age > MAX_PQ_AGE)
			lpq = remove_packet(lpq);
		last_pq = lpq;	
		lpq = lpq->next;
	}
	

	return ERR_SOCKNODATA; /* no data for socket */

rf_got_port:	


	/* Got a packet waiting */
	/* fill in sockaddr */
	__rt_memcpy(fromaddr, &list->s.si, sizeof(sockaddr));
	*addrlen = sizeof(sockaddr);
	__rt_memcpy(&fa->sin_addr.s_addr,&lpq->source_address,IP_ADDRESS_SIZE);
	__rt_memcpy(&fa->sin_port,&lpq->source_port,UDP_PORT_SIZE);
	
	/* Copy packet */
	packet_len = (len > lpq->size)?lpq->size:len;

	__rt_memcpy(buf, lpq->packet,packet_len);
	if (len < lpq->size) 	{	/* problem - not enough space for packet, remove from queue */
		framefree(lpq->frame_buffer);
		if (NULL == last_pq) /* was 1st entry in Q */
			pq = lpq->next;
   		else 
			last_pq->next = lpq->next;

		return ERR_SOCKBUFFERTOOSHORT; /* Packet too large */
	}

	/*
		Its fair to say that this packet is 'likely' to be from our host,
		so store this mac/ip pair in the 'safe' area. This is because the ARP
		takes a relatively long time to resolve an address
		Also there is no non-blocking here, ARPing would block on a send
		if it had to wait for a response (easy to fix non-blocking, just extra code
		which is not needed here).
		
	*/
	for (j=0;j<IP_ADDRESS_SIZE;j++) {
		ipaddr[j] = (unsigned char)(0xff & (fa->sin_addr.s_addr >> (j*8) ));
	}


	if (0 != arp_force_host_address(DEVICE, ipaddr, lpq->mac)) {
		// this error isn't critical-leave it		
	}
	
	/* Packet successfully read - remove from queue */
	remove_packet(lpq);
	return packet_len; /* Some data waiting */
}


/*
	Function: sendto(socket, msg, len, flags, destaddr, addrlen)
	
		socket - socket number
		msg - data of length len
		flags - unused here
		destaddr - struct of size addrlen 
	Description:
		Send msg to port and IP address described in destaddr
	Returns:
		0 - success
		-ve - fail
*/
  
int 	sendto(int socket, const void *msg, int len, int flags, const struct sockaddr *destaddr, int addrlen)
{
	inuse_socket_table *list;
	int offset = udp_data_offset();
	int err;
	unsigned char *fb, *packet, *bb;
	struct sockaddr_in da;


	if (addrlen == sizeof(sockaddr_in)) {
		__rt_memcpy(&da, (void*)destaddr, addrlen);
	}
	else {
		return -1;
	}
	// Find socket in table 
	if (NULL == (list = find_socket(socket))) {
		return ERR_SOCKSOCKNOTFOUND; /* socket not found */
	}
	
	
	// Is a frame large enough? 
	if ((len + offset) > ETHER_BUFFER_SIZE) {

		return ERR_SOCKPACKETTOOLONG; /* Packet too large */
	}
		
		
	// allocate memory for a frame 
	fb = framemalloc(ETHER_BUFFER_SIZE);
	if (NULL == fb) {
		return ERR_SOCKOUTOFFRAMES; // out of memory 
	}
		
	// calc offset for data 
	packet = fb + offset;
	
	// copy in data 
	__rt_memcpy(packet, msg, len);
	// Send it to (sockaddr_in)destaddr->sin_addr 
	err = udp_write(DEVICE, packet, (unsigned char*)&da.sin_port, 
			(unsigned char *)&da.sin_addr.s_addr, (unsigned char*)list->port_num, len, fb, &bb);

	/* free balance buffer - I don't want it */
	framefree(bb);

	return err;
}

/*
	func : new_sock_num(void)
	Desc: find an unused socket descriptor number This will eventually
	wrap around. Not a problem really
	return new socket number
*/
static int new_socket_num(void)
{
	static int i=1;
	inuse_socket_table *ist;
		
	ist = pIST;
	if (NULL != ist) 
		do {
			if (i == ist->socket_num) {
				i++;
				i &= ~(1 << (sizeof(int)-1)); /* Get rid of sign bit */
				ist = pIST; /* Restart */
			}
			ist = ist->next;
		} while (ist != NULL); 
		
	return i;
}


/*
	int socket (int af, int type, int protocol) 
	Params: 
		af - domain of communication, eg. internet AF_INET
		type - socket type, ie. comms type eg. SOCK_STREAM
		protocol - 	application level protocol, usually well known eg. bootp, tftp
	Returns:
		-ve on error
		+ve socket number on success- a reference used like a file handle.
	Restrictions: 
			only AF_INET supported
			no TCP hence only SOCK_DGRAM supported
			
*/
int	socket(int af, int type, int protocol)
{
	inuse_socket_table *ist, *last_ist;
	int sock;

	/* Find free table entry ie last entry */	
	last_ist = ist = pIST;	
    while (ist != NULL) {
		last_ist = ist;
		ist = ist->next;
	}  /* Find last entry */
   
	
		
	/* allocate space for new socket entry */
	ist = istmalloc(sizeof(inuse_socket_table));
	if (NULL == ist) {
		return ERR_SOCKOUTOFISTSPACE;
	}
	__rt_memset(ist,0,sizeof(inuse_socket_table));
	
 	sock = new_socket_num();
	
	if (NULL != last_ist)	/* There is more than 1 entry */
		last_ist->next = ist;
	else
		pIST = ist;	/* Was 1st entry, set the global static to point to new entry*/
		
	/* fill socket descriptor */
	ist->socket_num = sock;
	/* This is useful for selecting transmission type, UDP is the only
	one supported so it is ignored */
	ist->type = type;				/* Socket type (SOCK_DGRAM) */
	ist->s.sa.sa_family = af;		/* AF_INET only at present */
	/* return socket number */
	return sock;	
}


/*
	function: close(int socket)
	Desc: Find socket and free the memory
	returns:
		0 success,
		-VE error
*/

int close(int socket)
{
	inuse_socket_table *ist, *last_ist;
	
	last_ist = ist = pIST;		
	
	while (ist != NULL) {
		if (ist->socket_num == socket)
			goto found_socket;
		last_ist=ist;	
		ist = ist->next;	
	}
	return ERR_SOCKCLOSE;
		
found_socket:
	last_ist->next = ist->next;
	istfree(ist);
	
	return 0;		
}


unsigned short ntohs(unsigned short netshort) /* Convert network byte order to host order */
{
#ifdef LITTLE_ENDIAN
	netshort = ((0x00ff & (netshort >> 8)) | ((netshort << 8) & 0xff00));
#endif
return netshort;
}

unsigned int ntohl(unsigned int netlong) /* net long to host long */
{
#ifdef LITTLE_ENDIAN
	netlong=ntohs((U16)(netlong >> 16)) | (ntohs((U16)(netlong)) << 16);
#endif

return netlong;
}

unsigned short htons(unsigned short netshort) /* Convert network byte order to host order */
{
#ifdef LITTLE_ENDIAN
	netshort = ((0x00ff & (netshort >> 8)) | ((netshort << 8) & 0xff00));
#endif
return netshort;
}

unsigned int htonl(unsigned int netlong) /* net long to host long */
{
#ifdef LITTLE_ENDIAN
	netlong=ntohs((U16)(netlong >> 16)) | (ntohs((U16)(netlong)) << 16);
#endif

return netlong;
}

int nonblocking(int socket)
{
socket = socket;
 return 0;
}

#endif
/* EOF socket.c */

